/****************************
 * Calculate FFT with Signal
 ****************************/

(
~fft = {
	arg inputSignal, windowSize = 512;
	var complex, imag, cosTable;

	// Create a signal for the imaginary components of the input signal (required by fft)
	// Since we always use real signals, this will be an array of zeroes
	imag = Signal.newClear(windowSize);

	// Signal's fft requires a cosine table for efficient calculations
	cosTable = Signal.fftCosTable(windowSize);

	complex = fft(inputSignal, imag, cosTable);
	complex // Returns a complex number whose real part is a Signal and imag part is a Signal
};

~postFrequencyBins = {
	arg complexFreqs, windowSize, sampleRate;
	var mags = complexFreqs.magnitude, i = 0, threshold = 0.0001;

	while({i < (windowSize/2 + 1)}, {
		if(mags[i] > threshold, {
			var real = complexFreqs.real[i];
			var imag = complexFreqs.imag[i];

			post("Harmonic Number: ");
			post(i);
			post(", Frequency Bin: ");
			post(i/windowSize * sampleRate);
			post(", Amplitude: ");
			post(mags[i]*2/windowSize); // Note we need to scale by 2/N
			post(", Cosine Phase: ");
			postln(atan2(imag, real)) // atan2 takes arctan and accounts for quadrant

		});
		i = i + 1;
	});
}
)

(
// A simple periodic sine wave
var sampleRate = 44100; // Default server sample rate
var windowSize = 512;
var inputSignal1 = Signal.sineFill(windowSize, [0, 1], [0, pi/2]); // 2nd harmonic at phase pi/2
var complexFreqs1 = ~fft.(inputSignal1, windowSize); // periodic signal
"\nFrequencies for one cycle sine wave:".postln;
~postFrequencyBins.(complexFreqs1, windowSize, sampleRate);
inputSignal1.plot;

// IMPORTANT: Why don't we see the phase of pi/2 in printed output?  Because that's
// the phase of the sine wave and the fft reconstructs the phase of a cosine wave.
)

(
// A periodic square wave
var sampleRate = 44100; // Default server sample rate
var windowSize = 512;
var inputSignal2 = Signal.newClear(windowSize).waveFill(
	{arg x; if(x < (windowSize/2), {1}, {-1})},
	0,
	windowSize
);
var complexFreqs2 = ~fft.value(inputSignal2, windowSize); // periodic signal
"\nFrequencies for a periodic square wave:".postln;
~postFrequencyBins.value(complexFreqs2, windowSize, sampleRate);
)


/*********************
 * Using FFT and IFFT
 *********************/

(
SynthDef(\noProcMonoFFT, {
    arg out = 0;
	var in, chain, sig;
	in = Saw.ar(440, 0.2);
	chain = FFT(LocalBuf(2048), in);
	// Convert to frequency domain and convert back
	// No processing here
	// You can add processing here before you convert back
    sig = IFFT(chain);
	Out.ar(out, sig);
}).add;
)

x = Synth(\noProcMonoFFT);
x.free;

(
SynthDef(\noProcStereoFFT, {
    arg out = 0;
	var in, chain, sig, a;
	in = Saw.ar([440, 440], 0.2); // Stereo signal
	// Need two buffers for spectral data
	// Buffers should be passed in an array
	// Acceptable ways to do this:
	//     [LocalBuf(2048), LocalBuf(2048)]
	//     LocalBuf([2048, 2048])
	//     {LocalBuf(2048)} ! 2
	// All the above do the same thing.  Do not use
	// LocalBuf's multichannel argument.  FFT requires
	// separate buffer objects.
	chain = FFT({LocalBuf(2048)} ! 2, in);
	// Convert to frequency domain and convert back
	// No processing here
	// You can add processing here before you convert back
    sig = IFFT(chain);
	Out.ar(out, sig);
}).add;
)

x = Synth(\noProcStereoFFT);
x.free;

/********************
 * Load Audio File
 ******************/

(
~poker = Buffer.read(s, thisProcess.nowExecutingPath.dirname ++ "/pokerface.wav");
SynthDef(\play, {
	arg out = 0, bufnum;
	Out.ar(out, PlayBuf.ar(2, bufnum, doneAction: 2)); // loops!
}).add;
)

~playPoker = Synth(\play, [\bufnum, ~poker]);
~playPoker.free;

/********************
 * Brick Wall Filter
 ********************/

(
SynthDef(\brickWall, {
	arg out = 0, in, wipe = 0, bufnum;
	var chain, sig;
	in = PlayBuf.ar(2, bufnum, doneAction: 2);
	chain = FFT(LocalBuf([2048, 2048]), in); // Another way to expand to multichannel
	chain = PV_BrickWall(chain, wipe);
	// chain = PV_BrickWall(chain, SinOsc.kr(0.5) * 0.05);
	sig = IFFT(chain);
	Out.ar(out, sig);
}).add;
)

~brickWall = Synth(\brickWall, [\bufnum, ~poker]);
~brickWall.set(\wipe, 0.00);
~brickWall.free;

/**********************
 * Magnitude Smearing *
 **********************/

// Average a bin's magnitude with its neighbors

(
SynthDef(\magSmear, {
	arg out = 0, in, stretch = 1, shift = 0, interp = 0, bufnum;
	var chain, sig;
	in = PlayBuf.ar(2, bufnum, doneAction: 2);
	chain = FFT(LocalBuf([2048, 2048]), in);
	chain = PV_MagSmear(chain, MouseX.kr(0, 100));
	sig = IFFT(chain);
	Out.ar(out, sig);
}).add;
)
FreqScope.new
~magSmear = Synth(\binShift, [\bufnum, ~poker]);
~magSmear.free;


/************************
 * Bin Magnitude Freezing
 ************************/

// The magnitudes of the bins remain unchanged but changes in phase data
// allows the song to sound like it is still moving forward.  The result is
// a massively filtered effect that distorts the original song.
(
SynthDef(\magFreeze, {
	arg out = 0, in, bufnum, trigger = 0;
	var chain, sig;
	in = PlayBuf.ar(2, bufnum, loop: 1);
	chain = FFT(LocalBuf([512, 512]), in);
	chain = PV_Diffuser(chain, trigger);
	sig = IFFT(chain);
	Out.ar(out, sig);
}).add;
)

~freeze = Synth(\magFreeze, [\bufnum, ~poker]);
~freeze.set(\trigger, 1);
~freeze.set(\trigger, 0);
~freeze.free;

/******************
 * Time Stretching
 ******************/

// Mono Version

// First we need to determine a frame size (i.e., N) and a hop size.  Hop size is
// important for overlapping spectral data in playback.  Hop size of 0.25 is
// usually sufficient.

(
~frame = 4096;
~hop = 0.25;
~window_type = 0 // -1 = rectangle, 0 = sine, 1 = hann
)

// A mono recording of a piano sound.  We'll need to get several statistics about this
// file as well besides loading it into a buffer for playback.
//
// NOTE: you can't rely on a buffer's statistics being updated instantly.  Use the
// action method to perform some function once they are populated.
(
~filePath = thisProcess.nowExecutingPath.dirname ++ "/pianoMono.wav";
~piano = Buffer.read(s, ~filePath, action: {
	arg buf;
	~spectralBuf = Buffer.alloc(s, buf.duration.calcPVRecSize(~frame, ~hop), 1);
});
)

// MOTIVATION: A quick way to stretch is to playback a sound file at faster or
// slower speeds.  This changes pitch though!  Can we stretch without changing
// pitch?

(
SynthDef(\play, {
	arg out = 0, soundBufNum, rate = 1;
	var sig = PlayBuf.ar(1, soundBufNum, rate);
	Out.ar(out, sig ! 2);
}).add;
)

Synth(\play, [\soundBufNum, ~piano, \rate, 1]) // changing rate changes pitch

// Create spectral data in a buffer.  Wait for the recording to be complete
// if you plan to playback the data at a faster rate.

(
SynthDef(\pvrecMono, {
	arg spectralBufNum, soundBufNum;
	var in, chain;
	in = PlayBuf.ar(1, soundBufNum, 1, doneAction: 2);
	chain = FFT(LocalBuf(~frame), in, 0.25, ~window_type);
	chain =
	chain = PV_RecordBuf(chain, spectralBufNum, 0, 1, 0, ~hop, 1);
}).add;
)

// Create a buffer of spectral data
(
x = Synth(\pvrecMono, [\spectralBufNum, ~spectralBuf, \soundBufNum, ~piano]);
)
s.plotTree;

// Optionally save your spectral data to file!
~recBuf.write(~filePath ++ ".pvdata");

// Playback the spectral data
(
SynthDef(\pvplayMono, {
	arg out = 0, spectralBufNum, time;
	var in, chain;
	chain = PV_BufRd(LocalBuf(~frame), spectralBufNum, Line.kr(0, 1, time, doneAction: 2));
	Out.ar(out, IFFT(chain, ~window_type) ! 2);
}).add;
)

// Use the factor to determine the speed of playback.  For example, a factor of 2
// stretches the audio by 2.  Use a factor between 0 and 1 for faster playback.
(
var factor = 100;
Synth(\pvplayMono, [\spectralBufNum, ~spectralBuf, \time, ~piano.duration * factor]);
)

(
SynthDef(\pvplayMonoFreeze, {
	arg out = 0, spectralBufNum, location;
	var in, chain;
	chain = PV_BufRd(LocalBuf(~frame), spectralBufNum, location);
	Out.ar(out, IFFT(chain, ~window_type) ! 2);
}).add;
)

// Unlike PV_MagFreeze, we can now truly freeze a particular frame and play it back
// with overlap to create a smooth constant sound.
(
var location = 2.73; // 2.73 seconds into the sound file
Synth(\pvplayMonoFreeze, [\spectralBufNum, ~spectralBuf, \location, location/~piano.duration]);
)

/**************************
 * TIME STRETCHING STEREO *
 **************************/
(
~filePath = thisProcess.nowExecutingPath.dirname ++ "/pokerfaceChorus.wav";
~pokerL = Buffer.readChannel(s, ~filePath, channels: [0], action: {
	arg buf;
	~spectralBufL = Buffer.alloc(s, buf.duration.calcPVRecSize(~frame, ~hop), 1);
});
~pokerR = Buffer.readChannel(s, ~filePath, channels: [1], action: {
	arg buf;
	~spectralBufR = Buffer.alloc(s, buf.duration.calcPVRecSize(~frame, ~hop), 1);
});
)

(
SynthDef(\pvrecStereo, {
	arg spectralBufNumL, spectralBufNumR, soundBufNumL, soundBufNumR;
	var inL, inR, chainL, chainR;
	inL = PlayBuf.ar(1, soundBufNumL, 1);
	inR = PlayBuf.ar(1, soundBufNumR, 1, doneAction: 2);
	chainL = FFT(LocalBuf(~frame), inL, 0.25, ~window_type);
	chainR = FFT(LocalBuf(~frame), inR, 0.25, ~window_type);
	chainL = PV_RecordBuf(chainL, spectralBufNumL, 0, 1, 0, ~hop, 1);
	chainR = PV_RecordBuf(chainR, spectralBufNumR, 0, 1, 0, ~hop, 1);
}).add;
)

(
Synth(\pvrecStereo, [
	\spectralBufNumL, ~spectralBufL,
	\spectralBufNumR, ~spectralBufR,
	\soundBufNumL, ~pokerL,
	\soundBufNumR, ~pokerR
]);
)

s.plotTree;

(
SynthDef(\pvplayStereo, {
	arg out = 0, spectralBufNumL, spectralBufNumR, time;
	var chainL, chainR;
	chainL = PV_BufRd(LocalBuf(~frame), spectralBufNumL, Line.kr(0, 1, time));
	chainR = PV_BufRd(LocalBuf(~frame), spectralBufNumR, Line.kr(0, 1, time, doneAction: 2));
	Out.ar(out, [IFFT(chainL, ~window_type), IFFT(chainR, ~window_type)]);
}).add;
)

(
var factor = 4; // 100 is INSANITY
Synth(\pvplayStereo, [
	\spectralBufNumL, ~spectralBufL,
	\spectralBufNumR, ~spectralBufR,
	\time, ~pokerL.duration * factor
]);
)
